@@ -21,7 +21,7 @@
  */
 
 #include "rcs.h"
-RCS_ID("$Id: comm.c,v 1.8 1993/08/05 14:23:32 mlschroe Exp $ FAU")
+RCS_ID("$Id: comm.c,v 1.10 1993/11/30 19:28:07 mlschroe Exp $ FAU")
 
 #include "config.h"
 #include "acl.h"
@@ -45,28 +45,34 @@ struct comm comms[RC_LAST + 1] =
 #ifdef AUTO_NUKE
   { "autonuke",		NEED_DISPLAY|ARGS_ONE },
 #endif
-  { "bell",		ARGS_ONE },
+  { "bell",		ARGS_ZEROONE },
+  { "bell_msg",		ARGS_ZEROONE },
   { "bind",		ARGS_ONE|ARGS_ORMORE },
+#ifdef MAPKEYS
+  { "bindkey",		ARGS_ZERO|ARGS_ORMORE },
+#endif
   { "break",		NEED_FORE|ARGS_ZEROONE },
 #ifdef COPY_PASTE
   { "bufferfile",	ARGS_ZEROONE },
 #endif
+  { "c1",		ARGS_ZEROONE },
   { "chdir",		ARGS_ZEROONE },
   { "clear",		NEED_FORE|ARGS_ZERO },
 #ifdef MULTI
   { "clone",		NEED_DISPLAY|ARGS_ONE|ARGS_ORMORE },
 #endif
   { "colon",		NEED_DISPLAY|ARGS_ZERO },
+  { "command",		NEED_DISPLAY|ARGS_ZERO },
   { "console",		NEED_FORE|ARGS_ZEROONE },
 #ifdef COPY_PASTE
   { "copy",		NEED_FORE|ARGS_ZERO },
-  { "copy_reg",		ARGS_ZEROONE },
   { "crlf",		ARGS_ONE },
 #endif
   { "debug",		ARGS_ZEROONE },
 #ifdef AUTO_NUKE
   { "defautonuke",	ARGS_ONE },
 #endif
+  { "defescape",	ARGS_ONE },
   { "defflow",		ARGS_ONETWO },
 #if defined(UTMPOK) && defined(LOGOUTOK)
   { "deflogin",		ARGS_ONE },
@@ -78,16 +84,17 @@ struct comm comms[RC_LAST + 1] =
   { "defscrollback",	ARGS_ONE },
 #endif
   { "defwrap",		ARGS_ONE },
+  { "defwritelock",	ARGS_ONE },
   { "detach",		NEED_DISPLAY|ARGS_ZERO },
   { "displays",		NEED_DISPLAY|ARGS_ZERO },
   { "dumptermcap",	ARGS_ZERO },
-  { "duplicate",	ARGS_ZERO|ARGS_ORMORE },
   { "echo",		ARGS_ONETWO },
   { "escape",		ARGS_ONE },
 #ifdef PSEUDOS
   { "exec", 		NEED_FORE|ARGS_ZERO|ARGS_ORMORE },
 #endif
   { "flow",		NEED_FORE|ARGS_ZEROONE },
+  { "gr",		NEED_FORE|ARGS_ZEROONE },
   { "hardcopy",		NEED_FORE|ARGS_ZERO },
   { "hardcopy_append",	ARGS_ONE },
   { "hardcopydir",	ARGS_ONE },
@@ -98,8 +105,8 @@ struct comm comms[RC_LAST + 1] =
   { "history",		NEED_FORE|ARGS_ZERO },
 #endif
   { "info",		NEED_DISPLAY|ARGS_ZERO },
-#ifdef COPY_PASTE
-  { "ins_reg",		NEED_DISPLAY|ARGS_ZEROONE },
+#ifdef KANJI
+  { "kanji",		NEED_FORE|ARGS_ONETWO },
 #endif
   { "kill",		NEED_FORE|ARGS_ZERO },
   { "lastmsg",		NEED_DISPLAY|ARGS_ZERO },
@@ -107,11 +114,16 @@ struct comm comms[RC_LAST + 1] =
 #ifdef LOCK
   { "lockscreen",	NEED_DISPLAY|ARGS_ZERO },
 #endif
-  { "log",		ARGS_ZEROONE },
+  { "log",		NEED_FORE|ARGS_ZEROONE },
   { "logdir",		ARGS_ONE },
 #if defined(UTMPOK) && defined(LOGOUTOK)
   { "login",		NEED_FORE|ARGS_ZEROONE },
 #endif
+#ifdef MAPKEYS
+  { "mapdefault",	NEED_DISPLAY|ARGS_ZERO },
+  { "mapnotnext",	NEED_DISPLAY|ARGS_ZERO },
+  { "maptimeout",	ARGS_ZEROONE },
+#endif
 #ifdef COPY_PASTE
   { "markkeys",		ARGS_ONE },
 #endif
@@ -134,19 +146,22 @@ struct comm comms[RC_LAST + 1] =
   { "password",		ARGS_ZEROONE },
 #endif
 #ifdef COPY_PASTE
-  { "paste",		NEED_DISPLAY|ARGS_ZEROONE },
+  { "paste",		ARGS_ZEROONETWO },
+  { "pastefont",	ARGS_ZEROONE },
 #endif
   { "pow_break",	NEED_FORE|ARGS_ZEROONE },
 #ifdef POW_DETACH
   { "pow_detach",	NEED_DISPLAY|ARGS_ZERO },
-  { "pow_detach_msg",	ARGS_ONE },
+  { "pow_detach_msg",	ARGS_ZEROONE },
 #endif
   { "prev",		ARGS_ZERO },
+  { "printcmd",		ARGS_ZEROONE },
   { "process",		NEED_DISPLAY|ARGS_ZEROONE },
   { "quit",		ARGS_ZERO },
 #ifdef COPY_PASTE
   { "readbuf",		NEED_DISPLAY|ARGS_ZERO },
 #endif
+  { "readreg",          ARGS_ZEROONETWO },
   { "redisplay",	ARGS_ZERO },
   { "register",		ARGS_TWO },
 #ifdef COPY_PASTE
@@ -168,6 +183,7 @@ struct comm comms[RC_LAST + 1] =
   { "sleep",		ARGS_ONE },
   { "slowpaste",	ARGS_ONE },
   { "startup_message",	ARGS_ONE },
+  { "stuff",		NEED_DISPLAY|ARGS_ONE },
 #ifdef BSDJOBS
   { "suspend",		NEED_DISPLAY|ARGS_ZERO },
 #endif
@@ -178,7 +194,7 @@ struct comm comms[RC_LAST + 1] =
   { "title",		NEED_FORE|ARGS_ZEROONE },
   { "unsetenv",		ARGS_ONE },
   { "vbell",		ARGS_ZEROONE },
-  { "vbell_msg",	ARGS_ONE },
+  { "vbell_msg",	ARGS_ZEROONE },
   { "vbellwait",	ARGS_ONE },
   { "version",		ARGS_ZERO },
   { "wall",		NEED_DISPLAY|ARGS_ONE|ARGS_ORMORE },
